/*
								+---------------------------------+
								|                                 |
								|    ***   Bitmap font 2   ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2007  |
								|                                 |
								|          BitmapFont.h           |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 06-04-2007
 *
 *	changed bitmap font zoom style so when zoomed to 1 / pixel size, user gets perfectly crisp
 *	letters (in case font rendering functions are passed pixel aligned coordinates)
 *
 */

#ifndef __FONTENG_INCLUDED
#define __FONTENG_INCLUDED

class CGLFont {
public:
	/*
	 *	virtual bool Load(CGLState *p_state, const char *p_s_font_name)
	 *		- loads font from source identified as p_s_font_name
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Load(CGLState *p_state, const char *p_s_font_name) = 0;

	/*
	 *	virtual void Free()
	 *		- cleanup font, can't be used anymore (should be _ASSERT-ed)
	 */
	virtual void Free() = 0;

	/*
	 *	virtual void Zoom(float f_new_size = 1)
	 *		- scales font to f_new_size (not times f_new_size)
	 */
	virtual void Zoom(float f_new_size = 1) = 0;

	/*
	 *	virtual float f_Zoom() const
	 *		- returns current font scale (default 1)
	 */
	virtual float f_Zoom() const = 0;

	/*
	 *	virtual void Plot_Character(CGLState *p_state,
	 *		Vector2f v_pos, char n_char) const
	 *		- plots a single character n_char to position v_pos; p_state is OpenGL state guard
	 */
	virtual void Plot_Character(CGLState *p_state, Vector2f v_pos, char c_char) const = 0;

	/*
	 *	virtual void Plot_String(CGLState *p_state, Vector2f v_pos,
	 *		const char *p_s_string, int n_length = -1) const
	 *		- plots string of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used) to position v_pos; p_state is OpenGL state guard
	 *		- note this should be made faster than calling Plot_Character() in loop 
	 */
	virtual void Plot_String(CGLState *p_state, Vector2f v_pos,
		const char *p_s_string, int n_length = -1) const = 0;

	/*
	 *	virtual float f_StringWidth(const char *p_s_string, int n_length = -1) const
	 *		- returns width of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used)
	 *		- note in case string contains multiple lines of text, the longest line is returned
	 */
	virtual float f_StringWidth(const char *p_s_string, int n_length = -1) const = 0;

	/*
	 *	virtual Vector2f v_CharacterOffset(const char *p_s_string, int n_length = -1) const
	 *		- returns offset of n_length-th character from p_s_string (in case n_length is -1,
	 *		  returns offset of the last character)
	 *		- (note offset is vector pointing from first character of the string to the character)
	 */
	virtual Vector2f v_CharacterOffset(const char *p_s_string, int n_length = -1) const = 0;
};

#endif // __FONTENG_INCLUDED

#ifndef __BITMAP_FONT_INCLUDED
#define __BITMAP_FONT_INCLUDED

class CGLBitmapFont : public CGLFont {
protected:
	std::vector<TCharacterInfo> m_char_list;
	std::vector<CGLTexture_2D*> m_texture_list;
	unsigned char m_n_first_char;
	unsigned char m_n_char_num;
	int m_n_space_width;
	int m_n_newline_height;

	struct TCharLookup {
		GLuint n_display_list; // draw geometry, shift modelview. no glBegin() or glBindTexture()  
		const TCharacterInfo *p_character;
		CGLTexture_2D *p_tex_page;
	};

	TCharLookup *m_p_char_lookup;
	float m_f_zoom;

	class CDeleteObjects {
	public:
		template <class CObjType>
		inline void operator ()(CObjType *p_object) const
		{
			delete p_object;
		}
	};

	template <class CInnerObject>
	class CDeref {
	protected:
		CInnerObject m_object;

	public:
		inline CDeref(CInnerObject object)
			:m_object(object)
		{}

		template <class CObjType>
		inline void operator ()(CObjType *p_object)
		{
			m_object(*p_object); // dereference
		}
	};

	class CFindMin_Count {
	protected:
		unsigned char m_n_first_char;
		unsigned char m_n_char_num;
		unsigned char m_n_last_char;

	public:
		inline CFindMin_Count()
		{
			m_n_last_char = 0;
			m_n_first_char = 255;
		}

		inline unsigned char n_Min() { return m_n_first_char; }
		inline unsigned char n_Range() { return m_n_last_char - m_n_first_char + 1; }

		inline void operator ()(const TCharacterInfo &r_character_info)
		{
			if(r_character_info.n_code < m_n_first_char)
				m_n_first_char = r_character_info.n_code;
			if(r_character_info.n_code > m_n_last_char)
				m_n_last_char = r_character_info.n_code;
		}
	};

	class CFillCharLookup {
	protected:
		TCharLookup *m_p_char_lookup;
		CGLTexture_2D *m_p_cur_page;
		bool m_b_flip_texture;
		int m_n_newline_height;

	public:
		inline CFillCharLookup(TCharLookup *p_char_lookup, unsigned char n_first_char,
			CGLTexture_2D *p_cur_page, int n_newline_height, bool b_flip_texture = true)
			:m_p_char_lookup(p_char_lookup - n_first_char), m_p_cur_page(p_cur_page),
			m_n_newline_height(n_newline_height), m_b_flip_texture(b_flip_texture)
		{}

		inline void operator ()(const TCharacterInfo &r_character_info)
		{
			TCharLookup *p_cur_lookup = m_p_char_lookup + r_character_info.n_code;

			p_cur_lookup->p_character = &r_character_info;
			p_cur_lookup->p_tex_page = m_p_cur_page;

			p_cur_lookup->n_display_list = glGenLists(1);
			// create a new display list ...

			float f_tex_scale_u = 1.0f / m_p_cur_page->n_Width();
			float f_tex_scale_v = 1.0f / m_p_cur_page->n_Height();
			Vector2i v_min_texcoord = r_character_info.v_Min() - Vector2i(1, 1);
			Vector2i v_max_texcoord = r_character_info.v_Max(1);
			Vector2i v_min_vertex = r_character_info.v_baseline_offset - Vector2i(1, 1);
			Vector2i v_max_vertex = v_min_vertex + Vector2i(r_character_info.n_bitmap_width,
				r_character_info.n_bitmap_height) + Vector2i(1, 1);
			// create vertices and texcoords

			float f_jitter_u = -.5f * f_tex_scale_u;
			float f_jitter_v = -.5f * f_tex_scale_v;
			// crispy letters need sampling from texel centers

			glNewList(p_cur_lookup->n_display_list, GL_COMPILE);
			{
				int n_add = (m_b_flip_texture)? 1 : 0;
				int n_mul = (m_b_flip_texture)? -1 : 1;
				glTexCoord2f(v_min_texcoord.x * f_tex_scale_u + f_jitter_u,
					n_add + n_mul * v_min_texcoord.y * f_tex_scale_v + f_jitter_v);
				glVertex2f(v_min_vertex.x, v_min_vertex.y);
				glTexCoord2f(v_max_texcoord.x * f_tex_scale_u + f_jitter_u,
					n_add + n_mul * v_min_texcoord.y * f_tex_scale_v + f_jitter_v);
				glVertex2f(v_max_vertex.x, v_min_vertex.y);
				glTexCoord2f(v_max_texcoord.x * f_tex_scale_u + f_jitter_u,
					n_add + n_mul * v_max_texcoord.y * f_tex_scale_v + f_jitter_v);
				glVertex2f(v_max_vertex.x, v_max_vertex.y);
				glTexCoord2f(v_min_texcoord.x * f_tex_scale_u + f_jitter_u,
					n_add + n_mul * v_max_texcoord.y * f_tex_scale_v + f_jitter_v);
				glVertex2f(v_min_vertex.x, v_max_vertex.y);
				// draw aligned quad

				//glTranslatef(r_character_info.n_caret_shift, 0, 0);
				// translate past character to draw another one
				// apparently no matrix-ops in display lists :( (generates gl errors)
			}
			glEndList();
			// compile display list
		}
	};

public:
	/*
	 *	CGLBitmapFont::CGLBitmapFont()
	 *		- default constructor
	 */
	CGLBitmapFont();

	/*
	 *	CGLBitmapFont::~CGLBitmapFont()
	 *		- default destructor
	 */
	~CGLBitmapFont();

	/*
	 *	virtual bool CGLBitmapFont::Load(CGLState *p_state, const char *p_s_font_name)
	 *		- loads font from p_s_font_name bitmap (or more bitmaps)
	 *		- in case p_s_font_name ends with "_page00.tga", there must me at least
	 *		  one more page with characters (ie. "_page01.tga")
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Load(CGLState *p_state, const char *p_s_font_name);

#ifdef FONT_TEXTURE_CREATE_FONTS

	/*
	 *	virtual bool CGLBitmapFont::Create(const char *p_s_font_name, CGLState *p_state)
	 *		- loads font from p_s_font_name bitmap (or more bitmaps)
	 *		- in case p_s_font_name ends with "_page00.tga", there must me at least
	 *		  one more page with characters (ie. "_page01.tga")
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Create(CGLState *p_state, const char *p_s_font_face, int n_font_size,
		int n_font_weight = 400, bool b_italic = false, bool b_underline = false,
		int n_char_padding = 4);

#endif //FONT_TEXTURE_CREATE_FONTS

	/*
	 *	virtual void CGLBitmapFont::Free()
	 *		- cleanup font, can't be used anymore
	 */
	virtual void Free();

	/*
	 *	virtual void CGLBitmapFont::Zoom(float f_new_size = 1)
	 *		- zooms font to a given scale
	 */
	virtual void Zoom(float f_new_size = 1);

	/*
	 *	virtual float CGLBitmapFont::f_Zoom() const
	 *		- returns font scale
	 */
	virtual float f_Zoom() const;

	/*
	 *	virtual void CGLBitmapFont::Plot_Character(CGLState *p_state,
	 *		Vector2f v_pos, char n_char) const
	 *		- plots a single character n_char to position v_pos; p_state is OpenGL state guard
	 */
	virtual void Plot_Character(CGLState *p_state, Vector2f v_pos, char n_char) const;

	/*
	 *	virtual void CGLBitmapFont::Plot_String(CGLState *p_state, Vector2f v_pos,
	 *		const char *p_s_string, int n_length = -1) const
	 *		- plots string of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used) to position v_pos; p_state is OpenGL state guard
	 *		- note this is faster than calling Plot_Character() in loop can be 
	 */
	virtual void Plot_String(CGLState *p_state, Vector2f v_pos,
		const char *p_s_string, int n_length = -1) const;

	/*
	 *	virtual float CGLBitmapFont::f_StringWidth(const char *p_s_string, int n_length = -1) const
	 *		- returns width of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used)
	 *		- note in case string contains multiple lines of text, the longest line is returned
	 */
	virtual float f_StringWidth(const char *p_s_string, int n_length = -1) const;

	/*
	 *	virtual Vector2f CGLBitmapFont::v_CharacterOffset(const char *p_s_string,
	 *		int n_length = -1) const
	 *		- returns offset of n_length-th character from p_s_string (in case n_length is -1,
	 *		  it's the last character)
	 */
	virtual Vector2f v_CharacterOffset(const char *p_s_string, int n_length = -1) const;

protected:
	int n_Load_Page(const char *p_s_file_name, CGLState *p_state);
};

#endif //__BITMAP_FONT_INCLUDED
