/*
								+----------------------------------+
								|                                  |
								|    ***   Socket classes   ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2006   |
								|                                  |
								|            Socket.cpp            |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 03-25-2007
 *
 *	changed CStreamSocket::n_Read(), CStreamSocket::n_Write(), CDatagramSocket::n_SendTo()
 *	and CDatagramSocket::n_RecvFrom() to accept buffer size as unsigned int
 *
 *	passed code revision at 04-04-2007
 *
 *	added CSocket::GetPeerName() function to retrieve address of peer on the other side of socket
 *	renamed "TAddrStruct &TAddress::t_Address()" to "TAddrStruct &TAddress::r_Address()"
 *
 */

#ifdef WIN32
// the whole file is meant to be compiled under win32 only

#include <winsock.h>
#include "Socket.h"

/*
 *								=== CWinSock ===
 */

CWinSock::CWinSock()
	:m_b_ready(false)
{
	WORD n_min_version = MAKEWORD(2, 0);
	WSADATA t_winsock_info;

	if(WSAStartup(n_min_version, &t_winsock_info)) {
		WSACleanup();
		return;
	}

	if(t_winsock_info.wVersion != n_min_version) {
		WSACleanup();
		return;
	}
	// check WinSock version

	m_b_ready = true;
}

bool CWinSock::b_Ready()
{
	return m_b_ready;
}

CWinSock::~CWinSock()
{
	WSACleanup();
}

/*
 *								=== ~CWinSock ===
 */

/*
 *								=== TAddress ===
 */

/*
 *	TAddress::TAddress(const char *p_s_host_name, int n_port);
 *		- constructor with host name and port
 *		- note in case host name is 0, INADDR_ANY is used
 *		- in case host name is not 0, gethostbyname is called, which may fail
 *		  so the address validity flag is set accordingly (call b_Valid())
 */
TAddress::TAddress(const char *p_s_host_name, int n_port)
	:b_valid(false)
{
	t_address.sin_family = AF_INET;
	if(p_s_host_name) {
		LPHOSTENT p_host;
		if((p_host = gethostbyname(p_s_host_name)) == NULL)
			return;
		// host

		t_address.sin_addr = *((LPIN_ADDR)*p_host->h_addr_list);
	} else
		t_address.sin_addr.s_addr = INADDR_ANY;
	t_address.sin_port = htons(n_port);
	// address

	b_valid = true;
}

/*
 *								=== ~TAddress ===
 */

/*
 *								=== CSocket ===
 */

CWinSock CSocket::m_startup;

/*
 *	CSocket::CSocket()
 *		- default constructor, sets no socket
 */
CSocket::CSocket()
{
	m_n_socket = INVALID_SOCKET;
}

/*
 *	CSocket::~CSocket()
 *		- default destructor, releases socket in case it was created
 */
CSocket::~CSocket()
{
	if(m_n_socket != INVALID_SOCKET) {
		closesocket(m_n_socket);
		m_n_socket = INVALID_SOCKET;
	}
}

/*
 *	bool CSocket::b_State() const
 *		- returns true in case socket was succesfuly created, otherwise false
 */
bool CSocket::b_State() const
{
	return m_n_socket != INVALID_SOCKET;
}

/*
 *	bool CSocket::GetPeerName(TAddress &r_t_address) const
 *		- get address of peer on the other side of the socket
 *		- returns true on success, false on failure
 */
bool CSocket::GetPeerName(TAddress &r_t_address) const
{
	int n_addr_len = sizeof(struct sockaddr);
	return b_State() && getpeername(m_n_socket,
		(sockaddr*)&r_t_address.t_Address(), &n_addr_len) != SOCKET_ERROR;
}

/*
 *	bool CSocket::SetSockOpt(int n_level, int n_name, const void *p_value, int n_value_size)
 *		- sets socket options
 *		- returns true in case function succeeded, otherwise false
 *		- always returns false in case no socket was created
 */
bool CSocket::SetSockOpt(int n_level, int n_name, const void *p_value, int n_value_size)
{
	if(!b_State())
		return false;

	return !setsockopt(m_n_socket, n_level, n_name, (const char*)p_value, n_value_size);
}

/*
 *	bool CSocket::Close()
 *		- closes socket
 *		- in case socket was opened returns true. in case socket was closed already
 *		  (or has never been created) returns false
 */
bool CSocket::Close()
{
	if(!b_State())
		return false;
	closesocket(m_n_socket);
	m_n_socket = INVALID_SOCKET;
	return true;
}

/*
 *	bool CSocket::SetNonBlocking()
 *		- sets socket non-blocking
 *		- returns true in case it was succesfully completed, otherwise false
 *		- always returns false in case no socket was created
 *		- note in case one of p_Accept, Connect, n_Read or n_Write fails,
 *		  you have to check b_WouldBlock() to see wheter that was an error or
 *		  it was just cancelled in order not to block program execuiton (use with precaution!)
 */
bool CSocket::SetNonBlocking()
{
	if(!b_State())
		return false;
	unsigned long n_true = 1;
	if(ioctlsocket(m_n_socket, FIONBIO, &n_true) < 0) // todo - see if this param is correct 
		return false;
	return true;
}

/*
 *	bool CSocket::b_WouldBlock() const
 *		- returns true in case EWOULDBLOCK (or WSAEWOULDBLOCK under win) was set, otherwise false
 *		- this flag means that recently failed socket operation didn't fail because of real error
 *		  but because performing it would mean blocking program execution (non-blocking sockets)
 */
bool CSocket::b_WouldBlock() const
{
	return WSAGetLastError() == WSAEWOULDBLOCK;
}

/*
 *								=== ~CSocket ===
 */

/*
 *								=== CStreamSocket ===
 */

/*
 *	CStreamSocket::CStreamSocket()
 *		- default constructor, creates socket (call b_State() to determin wheter successfuly)
 */
CStreamSocket::CStreamSocket()
{
	m_n_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
}

/*
 *	bool CStreamSocket::Create()
 *		- creates socket
 *		- returns true in case socket is created in time of execution
 *		  leaving this function, otherwise false
 */
bool CStreamSocket::Create()
{
	if(b_State())
		return true;
	return (m_n_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) != INVALID_SOCKET;
}

/*
 *	bool CStreamSocket::Connect(const TAddress &r_t_address)
 *		- attempts to connect to host given by r_t_address
 *		- returns true in case connection was established, otherwise false
 *		- always returns false in case no socket was created
 */
bool CStreamSocket::Connect(const TAddress &r_t_address)
{
	if(!b_State())
		return false;

	if(connect(m_n_socket, (const sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr)) == SOCKET_ERROR)
		return false;
	return true;
}

/*
 *	bool CStreamSocket::Listen(const TAddress &r_t_address, int n_max_connections)
 *		- binds socket to address r_t_address, begin listening for n_max_connections
 *		- returns true in case socket is listening, otherwise false
 *		- always returns false in case no socket was created
 */
bool CStreamSocket::Listen(const TAddress &r_t_address, int n_max_connections)
{
	if(!b_State())
		return false;

	if(bind(m_n_socket, (const sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr)) == SOCKET_ERROR ||
	   listen(m_n_socket, n_max_connections) == SOCKET_ERROR)
		return false;
	return true; 
}

/*
 *	CStreamSocket *CStreamSocket::p_Accept(TAddress &r_t_address) const
 *		- accepts incoming connections; r_t_address is filled with network address of socket
 *		  which is returned
 *		- returns a new socket, connected to the other side or 0 on failure
 *		- note it propably will return 0 in case Listen() was not successfuly called prior calling this function
 *		- always returns 0 in case no socket was created
 */
CStreamSocket *CStreamSocket::p_Accept(TAddress &r_t_address) const
{
	SOCKET n_socket;
	int n_addr_len = sizeof(struct sockaddr);
	if((n_socket = accept(m_n_socket, (sockaddr*)&r_t_address.t_Address(), &n_addr_len)) == INVALID_SOCKET)
		return 0;
	// maybe just timeout or something ... do not disconnect me yet

	CStreamSocket *p_new;
	if(!(p_new = new CStreamSocket))
		return 0;
	if(p_new->b_State())
		closesocket(p_new->m_n_socket); // close it right away
	p_new->m_n_socket = n_socket;
	// assign socket

	return p_new;
}

/*
 *	int CStreamSocket::n_Read(void *p_buffer, unsigned int n_buffer_size) const
 *		- receives data from socket to buffer
 *		- maximum amount of data read is given by n_buffer_size but it can also read less data
 *		- returns number of bytes written into the buffer on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case there are no data for the socket, the function is by default waiting
 *		  until some arrives (can be overriden by calling SetNonBlocking())
 */
int CStreamSocket::n_Read(void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = recv(m_n_socket, (char*)p_buffer, n_buffer_size, 0);
	if(n_result == SOCKET_ERROR)
		return -1;
	return n_result;
}

/*
 *	int CStreamSocket::n_Write(const void *p_buffer, unsigned int n_buffer_size) const
 *		- transmits n_buffer_size bytes from buffer to socket
 *		- returns number of bytes transmitted into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CStreamSocket::n_Write(const void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = send(m_n_socket, (const char*)p_buffer, n_buffer_size, 0);
	if(n_result == SOCKET_ERROR)
		return -1;
	return n_result;
}

/*
 *								=== ~CStreamSocket ===
 */

/*
 *								=== CDatagramSocket ===
 */

/*
 *	CDatagramSocket::CDatagramSocket()
 *		- default constructor, creates socket (call b_State() to determin wheter successfuly)
 */
CDatagramSocket::CDatagramSocket()
{
	m_n_socket = socket(AF_INET, SOCK_DGRAM, 0);
}

/*
 *	bool CDatagramSocket::Create()
 *		- creates socket
 *		- returns true in case socket is created in time of execution
 *		  leaving this function, otherwise false
 */
bool CDatagramSocket::Create()
{
	if(b_State())
		return true;
	return (m_n_socket = socket(AF_INET, SOCK_DGRAM, 0)) != INVALID_SOCKET;
}

/*
 *	bool CDatagramSocket::Bind(const TAddress &r_t_address)
 *		- binds the socket to address r_t_address
 *		- returns true on succes or false on failure
 *		- always returns false in case no socket was created
 */
bool CDatagramSocket::Bind(const TAddress &r_t_address)
{
	if(!b_State())
		return false;

	if(bind(m_n_socket, (const sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr)) == SOCKET_ERROR)
		return false;
	return true;
}

/*
 *	int CDatagramSocket::n_SendTo(const TAddress &r_t_address,
 *		const void *p_buffer, unsigned int n_buffer_size) const
 *		- sends n_buffer_size bytes to  socket from buffer
 *		- returns number of bytes written into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CDatagramSocket::n_SendTo(const TAddress &r_t_address,
	const void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = sendto(m_n_socket, (const char*)p_buffer, n_buffer_size, 0,
		(const sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr));
	if(n_result == SOCKET_ERROR)
		return -1;
	return n_result;
}

/*
 *	int CDatagramSocket::n_RecvFrom(const TAddress &r_t_address,
 *		void *p_buffer, unsigned int n_buffer_size) const
 *		- receives n_buffer_size bytes from socket to buffer; r_t_address is set to address which sent the data
 *		- returns number of bytes written into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CDatagramSocket::n_RecvFrom(TAddress &r_t_address,
	void *p_buffer, unsigned int n_buffer_size) const
{
	int n_addr_len = sizeof(struct sockaddr);
	int n_result = recvfrom(m_n_socket, (char*)p_buffer, n_buffer_size, 0,
		(sockaddr*)&r_t_address.t_Address(), &n_addr_len);
	if(n_result == SOCKET_ERROR)
		return -1;
	return n_result;
}

/*
 *								=== ~CDatagramSocket ===
 */

#else //WIN32
// the whole file is meant to be compiled under linux only

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include "Socket.h"

/*
 *								=== TAddress ===
 */

/*
 *	TAddress::TAddress(const char *p_s_host_name, int n_port);
 *		- constructor with host name and port
 *		- note in case host name is 0, INADDR_ANY is used
 *		- in case host name is not 0, gethostbyname is called, which may fail
 *		  so the address validity flag is set accordingly (call b_Valid())
 */
TAddress::TAddress(const char *p_s_host_name, int n_port)
	:b_valid(false)
{
	t_address.sin_family = AF_INET;
	t_address.sin_port = htons(n_port);
	if(p_s_host_name) {
		struct hostent *p_host;
		if((p_host = gethostbyname(p_s_host_name)) == NULL)
			return;
		memcpy(&t_address.sin_addr, p_host->h_addr, p_host->h_length);
	} else
		t_address.sin_addr.s_addr = INADDR_ANY;
	b_valid = true;
}

/*
 *								=== ~TAddress ===
 */

/*
 *								=== CSocket ===
 */

/*
 *	CSocket::CSocket()
 *		- default constructor, sets no socket
 */
CSocket::CSocket()
{
	m_n_socket = -1;
}

/*
 *	CSocket::~CSocket()
 *		- default destructor, releases socket in case it was created
 */
CSocket::~CSocket()
{
	if(m_n_socket >= 0) {
		close(m_n_socket);
		m_n_socket = -1;
	}
}

/*
 *	bool CSocket::b_State() const
 *		- returns true in case socket was succesfuly created, otherwise false
 */
bool CSocket::b_State() const
{
	return m_n_socket >= 0;
}

/*
 *	bool CSocket::GetPeerName(TAddress &r_t_address) const
 *		- get address of peer on the other side of the socket
 *		- returns true on success, false on failure
 */
bool CSocket::GetPeerName(TAddress &r_t_address) const
{
	unsigned int n_addr_len = sizeof(struct sockaddr_in);
	return b_State() && !getpeername(m_n_socket,
		(struct sockaddr*)&r_t_address.t_Address(), &n_addr_len);
}

/*
 *	bool CSocket::SetSockOpt(int n_level, int n_name, const void *p_value, int n_value_size)
 *		- sets socket options
 *		- returns true in case function succeeded, otherwise false
 *		- always returns false in case no socket was created
 */
bool CSocket::SetSockOpt(int n_level, int n_name, const void *p_value, int n_value_size)
{
	if(!b_State())
		return false;

	return !setsockopt(m_n_socket, n_level, n_name, p_value, n_value_size);
}

/*
 *	bool CSocket::Close()
 *		- closes socket
 *		- in case socket was opened returns true. in case socket was closed already
 *		  (or has never been created) returns false
 */
bool CSocket::Close()
{
	if(!b_State())
		return false;
	close(m_n_socket);
	m_n_socket = -1;
	return true;
}

/*
 *	bool CSocket::SetNonBlocking()
 *		- sets socket non-blocking
 *		- returns true in case it was succesfully completed, otherwise false
 *		- always returns false in case no socket was created
 *		- note in case one of p_Accept, Connect, n_Read or n_Write fails,
 *		  you have to check b_WouldBlock() to see wheter that was an error or
 *		  it was just cancelled in order not to block program execuiton (use with precaution!)
 */
bool CSocket::SetNonBlocking()
{
	if(!b_State())
		return false;
	if(fcntl(m_n_socket, F_SETFL, O_NDELAY) < 0)
		return false;
	return true;
}

/*
 *	bool CSocket::b_WouldBlock() const
 *		- returns true in case EWOULDBLOCK (or WSAEWOULDBLOCK under win) was set, otherwise false
 *		- this flag means that recently failed socket operation didn't fail because of real error
 *		  but because performing it would mean blocking program execution (non-blocking sockets)
 */
bool CSocket::b_WouldBlock() const
{
	return errno == EWOULDBLOCK;
}

/*
 *								=== ~CSocket ===
 */

/*
 *								=== CStreamSocket ===
 */

/*
 *	CStreamSocket::CStreamSocket()
 *		- default constructor, creates socket (call b_State() to determin wheter successfuly)
 */
CStreamSocket::CStreamSocket()
{
	m_n_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
}

/*
 *	bool CStreamSocket::Create()
 *		- creates socket
 *		- returns true in case socket is created in time of execution
 *		  leaving this function, otherwise false
 */
bool CStreamSocket::Create()
{
	if(b_State())
		return true;
	return (m_n_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) >= 0;
}

/*
 *	bool CStreamSocket::Connect(const TAddress &r_t_address)
 *		- attempts to connect to host given by r_t_address
 *		- returns true in case connection was established, otherwise false
 *		- always returns false in case no socket was created
 */
bool CStreamSocket::Connect(const TAddress &r_t_address)
{
	if(!b_State())
		return false;

	if(connect(m_n_socket, (const struct sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr_in)) < 0)
		return false;
	return true;
}

/*
 *	bool CStreamSocket::Listen(const TAddress &r_t_address, int n_max_connections)
 *		- binds socket to address r_t_address, begin listening for n_max_connections
 *		- returns true in case socket is listening, otherwise false
 *		- always returns false in case no socket was created
 */
bool CStreamSocket::Listen(const TAddress &r_t_address, int n_max_connections)
{
	if(!b_State())
		return false;

	if(bind(m_n_socket, (const struct sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr_in)) < 0 ||
	   listen(m_n_socket, n_max_connections)) // fails with nonzero
		return false;
	return true; 
}

/*
 *	CStreamSocket *CStreamSocket::p_Accept(TAddress &r_t_address) const
 *		- accepts incoming connections; r_t_address is filled with network address of socket
 *		  which is returned
 *		- returns a new socket, connected to the other side or 0 on failure
 *		- note it propably will return 0 in case Listen() was not successfuly called prior calling this function
 *		- always returns 0 in case no socket was created
 */
CStreamSocket *CStreamSocket::p_Accept(TAddress &r_t_address) const // accept incoming connections on this address
{
	int n_socket;
    unsigned int n_length = sizeof(struct sockaddr_in);
	if((n_socket = accept(m_n_socket, (struct sockaddr*)&r_t_address.t_Address(),
	   &n_length)) < 0)
		return 0;
	// maybe just timeout or something ... do not disconnect me yet

	CStreamSocket *p_new;
	if(!(p_new = new CStreamSocket))
		return 0;
	if(p_new->b_State())
		close(p_new->m_n_socket); // close it right away
	p_new->m_n_socket = n_socket;
	// assign socket

	return p_new;
}

/*
 *	int CStreamSocket::n_Read(void *p_buffer, unsigned int n_buffer_size) const
 *		- receives data from socket to buffer
 *		- maximum amount of data read is given by n_buffer_size but it can also read less data
 *		- returns number of bytes written into the buffer on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case there are no data for the socket, the function is by default waiting
 *		  until some arrives (can be overriden by calling SetNonBlocking())
 */
int CStreamSocket::n_Read(void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = read(m_n_socket, (char*)p_buffer, n_buffer_size);
	if(n_result < 0)
		return -1;
	return n_result;
}

/*
 *	int CStreamSocket::n_Write(const void *p_buffer, unsigned int n_buffer_size) const
 *		- transmits n_buffer_size bytes from buffer to socket
 *		- returns number of bytes transmitted into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CStreamSocket::n_Write(const void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = write(m_n_socket, (const char*)p_buffer, n_buffer_size);
	if(n_result < 0)
		return -1;
	return n_result;
}

/*
 *								=== ~CStreamSocket ===
 */

/*
 *								=== CDatagramSocket ===
 */

/*
 *	CDatagramSocket::CDatagramSocket()
 *		- default constructor, creates socket (call b_State() to determin wheter successfuly)
 */
CDatagramSocket::CDatagramSocket()
{
	m_n_socket = socket(AF_INET, SOCK_DGRAM, 0);
}

/*
 *	bool CDatagramSocket::Create()
 *		- creates socket
 *		- returns true in case socket is created in time of execution
 *		  leaving this function, otherwise false
 */
bool CDatagramSocket::Create()
{
	if(b_State())
		return true;
	return (m_n_socket = socket(AF_INET, SOCK_DGRAM, 0)) >= 0;
}

/*
 *	bool CDatagramSocket::Bind(const TAddress &r_t_address)
 *		- binds the socket to address r_t_address
 *		- returns true on succes or false on failure
 *		- always returns false in case no socket was created
 */
bool CDatagramSocket::Bind(const TAddress &r_t_address)
{
	if(!b_State())
		return false;

	if(bind(m_n_socket, (struct sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr_in)) < 0)
		return false;
	return true; 
}

/*
 *	int CDatagramSocket::n_SendTo(const TAddress &r_t_address,
 *		const void *p_buffer, unsigned int n_buffer_size) const
 *		- sends n_buffer_size bytes to  socket from buffer
 *		- returns number of bytes written into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CDatagramSocket::n_SendTo(const TAddress &r_t_address,
	const void *p_buffer, unsigned int n_buffer_size) const
{
	int n_result = sendto(m_n_socket, (const char*)p_buffer, n_buffer_size, 0,
		(const sockaddr*)&r_t_address.t_Address(), sizeof(struct sockaddr));
	if(n_result < 0)
		return -1;
	return n_result;
}

/*
 *	int CDatagramSocket::n_RecvFrom(const TAddress &r_t_address,
 *		void *p_buffer, unsigned int n_buffer_size) const
 *		- receives n_buffer_size bytes from socket to buffer; r_t_address is set to address which sent the data
 *		- returns number of bytes written into socket on success or -1 on failure
 *		- always returns -1 in case no socket was created
 *		- note in case the socket buffer is full or there is another reason why the
 *		  operation, canot be performed immediately the function is by default waiting
 *		  until it can proceed (can be overriden by calling SetNonBlocking())
 */
int CDatagramSocket::n_RecvFrom(TAddress &r_t_address,
	void *p_buffer, unsigned int n_buffer_size) const
{
	unsigned int n_addr_len = sizeof(struct sockaddr);
	int n_result = recvfrom(m_n_socket, (char*)p_buffer, n_buffer_size, 0,
		(sockaddr*)&r_t_address.t_Address(), &n_addr_len);
	if(n_result < 0)
		return -1;
	return n_result;
}

/*
 *								=== ~CDatagramSocket ===
 */

#endif //WIN32
